/* =========================
   P31_D88
   by Koichi Nishida
   (c)2013 Nishida Radio
==========================*/

#include <stdio.h>
#include <windows.h>
#include <shlwapi.h>
#include <tchar.h>
#pragma comment(lib, "shlwapi.lib")

// create d88 from pp31
void p31_d88(FILE *fp_p31, FILE *fp_d88, char *name);

char *StrToUpper(char *s)
{
    char *p;  
 
    for (p = s; *p; p++) *p = toupper(*p);
	return s;
}

void conv(LPCTSTR name)
{
	if (!name || !name[0]) return;

	FILE *fp_p31,  *fp_d88;
	TCHAR path[MAX_PATH], d88name[MAX_PATH];

	if (StrCmp(StrToUpper(PathFindExtension(name)), TEXT(".P31")) != 0) {
		MessageBox(NULL, TEXT("Drop .P31 file."), TEXT("P31_D88"), MB_OK);
		return;
	}
	if (fopen_s(&fp_p31, (char *)name, "rb")) {
		MessageBox(NULL, TEXT("Can't open the P31 file."), TEXT("P31_D88"), MB_OK);
		return;
	}

	lstrcpy(path, name);
	PathRemoveExtension(path);
	lstrcpy(d88name, path);
	PathAddExtension(path, TEXT(".D88"));
	if (fopen_s(&fp_d88, (char *)path, "wb")) {
		MessageBox(NULL, TEXT("Can't create D88 file."), TEXT("P31_D88"), MB_OK);
		fclose(fp_p31);
		return;
	}

	p31_d88(fp_p31, fp_d88, d88name);

	MessageBox(NULL, TEXT("the D88 image created."), TEXT("P31_D88"), MB_OK);
	fclose(fp_d88);
	fclose(fp_p31);
}

// create d88 file from p31
void p31_d88(FILE *fp_p31, FILE *fp_d88, char *name)
{	
	int trk, sec, i;
	unsigned long dsize;

	// d88 header
	for (i=0; i<16; i++) fputc(0, fp_d88);
	fputc(0, fp_d88);
	for (i=0; i<9; i++) fputc(0, fp_d88);	// reserved
	fputc(0x00, fp_d88);					// write protect no
	fputc(0x00, fp_d88);					// disk type = 2D
	dsize = 32+4*164+(16+256)*16*40;		// disk size	
	fputc(dsize&0xff, fp_d88);
	fputc((dsize>>8)&0xff, fp_d88);
	fputc((dsize>>16)&0xff, fp_d88);
	fputc((dsize>>24)&0xff, fp_d88);

	// pointers to tracks
	// 0 - 79 track
	for (trk=0; trk<80; trk++) {
		if (trk & 1) {
			fputc(0, fp_d88);
			fputc(0, fp_d88);
			fputc(0, fp_d88);
			fputc(0, fp_d88);
		}
		else {
			unsigned long val = 32 + 4 * 164 + (16 + 256) * /*16*/8 * trk;
			fputc(val & 0xff, fp_d88);
			fputc((val >> 8) & 0xff, fp_d88);
			fputc((val >> 16) & 0xff, fp_d88);
			fputc((val >> 24) & 0xff, fp_d88);
		}
	}
	// 80 - 163 track

	for (trk=80; trk<164; trk++) {
		fputc(0, fp_d88);
		fputc(0, fp_d88);
		fputc(0, fp_d88);
		fputc(0, fp_d88);
	}

	// track data
	for (trk=0; trk<80; trk++) {
		if ((trk&1)==0) {
			for (sec = 1; sec < 17; sec++) {
				// sector header
				fputc(trk >> 1, fp_d88);					// C (cylinder)
				fputc(trk & 1, fp_d88);					// H (side)
				fputc(sec, fp_d88);						// R (sector)
				fputc(1, fp_d88);						// N (256 bytes/sector)
				fputc(16, fp_d88);						// 16 sectors/track
				fputc(0, fp_d88);
				fputc(0, fp_d88);						// double density
				fputc(0, fp_d88);						// deleted mark
				fputc(0, fp_d88);						// status
				for (i = 0; i < 5; i++) fputc(0, fp_d88);	// reserved
				fputc(0, fp_d88); fputc(1, fp_d88);	// data size of sector part
				{
					unsigned char buff[256];

					fread(buff, 1, 256, fp_p31);
					fwrite(buff, 1, 256, fp_d88);
					fread(buff, 1, 256, fp_p31);
				}
			}
		} else {
			for (sec = 1; sec < 17; sec++) {
				unsigned char buff[256];
				fread(buff, 1, 256, fp_p31);
				fread(buff, 1, 256, fp_p31);
			}
		}
	}
}

int APIENTRY WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
                     LPSTR lpCmdLine, int nCmdShow)
{
	conv(__targv[1]);

	return 0;
}