﻿/*
 * INI.c
 *
 * Created: 2013/11/06 1:16:33
 *  Author: Koichi Nishida
 */ 
/*
Copyright (C) 2015 Koichi NISHIDA
email to Koichi NISHIDA: tulip-house@msf.biglobe.ne.jp
*/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <string.h>
#include "INI.h"
#include "SD.h"
#include "COMMON.h"
#include "LCD.h"

// SD6031AR.INI file
struct FILE iniFile;

// open or create SD6031AR.INI file
void INI_openCreate(uint8_t *err)
{
	FILE_open(&iniFile, 0, "SD6031AR", "INI", 0, err);
	if (*err) {
		*err = 0;
		if (WP) {
#ifdef AIRDISP
			LCD_printAll("Write   Protect.");
#endif
		} else {
			FILE_create(0, "SD6031ARINI", 512, err);
			if (*err) return;
			FILE_open(&iniFile, 0, "SD6031AR", "INI", 0, err);
			if (*err) return;

			FILE_writeBegin(&iniFile, 0, err);	
			for (uint8_t j = 0; j < 4; j++) {
				SPI_writeByte(0, err);
				if (*err) return;
				for (uint16_t i = 1; i < 64; i++) {
					SPI_writeByte(' ', err);
					if (*err) return;
				}
			}
			for (uint16_t j = 0; j < 256; j++) {
				SPI_writeByte(' ', err);
				if (*err) return;
			}
			FILE_writeEnd(err);
		}
	}
}

// read INI file to a buffer
void INI_read(uint8_t *buff, uint8_t *err)
{
	FILE_readBegin(&iniFile, 0, err);
	if (*err) return;
	for (uint16_t i=0; i<512; i++) {
		buff[i] = SPI_readByte(err);
		if (*err) return;
	}
	FILE_readEnd(err);
}

// write INI file
void INI_write(uint8_t *buff, uint8_t *err)
{
	FILE_writeBegin(&iniFile, 0, err);
	if (*err) return;
	for (uint16_t i=0; i<512; i++) {
		SPI_writeByte(buff[i], err);
		if (*err) return;
	}
	FILE_writeEnd(err);
}

// read a file name in SD6031AR.INI
void INI_readFName(uint8_t *buff, uint8_t drv, char *name, uint8_t *err)
{
	INI_read(buff, err);
	if (*err) return;
	strcpy(name, (char *)buff+(uint16_t)drv*64);
}

// substitute a file name in SD6031AR.INI
void INI_substitute(uint8_t *buff, uint8_t drv, char *name, uint8_t *err)
{
	INI_read(buff, err);
	if (*err) return;
	for (uint8_t i=0; i<64; i++) {
		buff[(uint16_t)drv*64+i]=' ';
	}
	strcpy((char *)buff+(uint16_t)drv*64, name);
	INI_write(buff, err);
}
