/*======================================================

	SD6031Air Firmware
	
	2016.3.28 by Koichi Nishida
	Copyright (C) 2016 Koichi NISHIDA
	email : tulip-house@msf.biglobe.ne.jp

======================================================*/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <string.h>
#include "SD.h"
#include <util/delay.h>
#include "COMMON.h"
#include "BUFFER.h"
#include "COMU.h"
#include "LCD.h"
#include "FLASHAIR.h"
#include "MOUNT.h"

#define WAIT 1
#define min(a,b) ((a)>(b)?(b):(a))

// ==================== prototypes ====================

// calculate 2^x
uint8_t pow2(register uint8_t x);

#define DRIVE_NUM 4

// send P31 image file names to P6
void sendFileNames(uint8_t drive);
	
// ---------- communication with P6  ----------
// communicate with PC-6001
// send data to PC-6001
void P6_snd(uint8_t d1, uint8_t d2);
// receive data to PC-6001
void P6_rcv(uint8_t *d1, uint8_t *d2);

// send double data to PC-6001
void P6_sndw_pre(void);
void P6_sndw(uint8_t d1, uint8_t d2, uint8_t d3, uint8_t d4);
void P6_sndw_post(void);
// receive double data to PC-6001
void P6_rcvw_pre(void);
void P6_rcvw(uint8_t *d1, uint8_t *d2, uint8_t *d3, uint8_t *d4);
void P6_rcvw_post(void);

// ---------- PC-6031 status ----------
struct pc6031_status {
	uint8_t drive;
	uint16_t track;
	uint8_t sector;
	uint8_t length;
	uint8_t valid;
	uint8_t mode;
	uint8_t startSec[DRIVE_NUM];
	uint8_t endSec[DRIVE_NUM];
	uint8_t error;
};

// initialize PC-6031 status
void pc6031_init(void);

// run PC-6031
void pc6031_process();

//==================== implementation ====================

uint8_t globalProtect[DRIVE_NUM];

struct pc6031_status pc6031;	// PC-6031 status
uint8_t LED;				// LED status
uint8_t RST_init;

// calculate 2^x
uint8_t pow2(register uint8_t x)
{
	uint8_t i, r=1;

	for (i=0; i!=x; i++) {
		r<<=1;
	}
	return r;
}

void sendAFileName(struct FILELST *list, uint16_t *entryNum, uint32_t blkadr, uint16_t ofs)
{
	uint8_t err = 0;
	
	SD_readBlock(blkadr, &err);
	if (err) return;
	P6_snd(1, 0);
	for (uint8_t i=0; i<8; i++) P6_snd(SD_p.buff[ofs+i],0);
}

// send P31 image file names to P6
void sendFileNames(uint8_t drive)
{
	uint8_t err = 0;	
		
	if (EJECT) return;
	
	struct FILE directory;
	char ext_[4]="   ", name_[9]="D       ";
	name_[1] = ('1'+drive);
	FILE_open(&directory, 0, name_, ext_, 0, &err);
	if (err) return;
	struct FILELST list;
	uint16_t entryNum;
	
	FILE_traverseADirectory(directory.startCluster, "P31", &list, &entryNum, &sendAFileName, &err);
	if (err) return;

	P6_snd(0, 0);
}

// communicate with PC-6001
// send data to PC-6001
void P6_snd(uint8_t d1, uint8_t d2)
{ 
	while (!(COMU_IN_C()&PIN5_bm)) ;	// wait RFD set

	COMU_OUT_A(d1);
	
    COMU_OUT_C(PIN0_bm);				// set DAV

	while (!(COMU_IN_C()&PIN6_bm)) ;	// wait DAC set

	COMU_OUT_A(d2);
	
	COMU_OUT_C(0);						// reset DAV
	
	while (COMU_IN_C()&PIN6_bm) ;		// wait DAC reset
}

// receive data from PC-6001
void P6_rcv(uint8_t *d1, uint8_t *d2)
{
	*d1 = 0xff;
	if (d2) *d2 = 0xff;

	COMU_OUT_C(PIN1_bm);				// set RFD

	while (!(COMU_IN_C()&PIN4_bm)) ;	// wait DAV set

	COMU_OUT_C(0);						// reset RFD

	*d1 = COMU_IN_B();

	COMU_OUT_C(PIN2_bm);				// set DAC

	while (COMU_IN_C()&PIN4_bm) ;		// wait DAV reset

	if (d2) *d2 = COMU_IN_B();

	COMU_OUT_C(0);						// reset DAC
}

// pre-send double data to PC-6001
void P6_sndw_pre(void)
{ 
	// while (bit_is_clear(PIND,1)) ;	// wait RFD set
	while (!(COMU_IN_C()&PIN5_bm)) ;	// wait RFD set

	// DDRA  = 0b11111111;				// data out to P6
	COMU_DIR_A(0b11111111);
	
	// DDRC  = 0b11111111;				// data out to P6
	COMU_DIR_B(0b11111111);
}

// send double data to PC-6001
void P6_sndw(uint8_t d1, uint8_t d2, uint8_t d3, uint8_t d4)
{ 
	// PORTA = d1;
	// PORTC = d2;
	COMU_OUT_A(d1);
	COMU_OUT_B(d2);

	// PORTD |= 0b00010000;				// set DAV
	COMU_OUT_C(PIN0_bm);

	// while (bit_is_clear(PIND,2)) ;	// wait DAC set
	while (!(COMU_IN_C()&PIN6_bm)) ;	

	// PORTA = d3;
	// PORTC = d4;
	COMU_OUT_A(d3);
	COMU_OUT_B(d4);

	// PORTD &= 0b11101111;			// reset DAV
	COMU_OUT_C(0);
	
	// while (bit_is_set(PIND,2)) ;	// wait DAC reset
	while (COMU_IN_C()&PIN6_bm) ;
}

// post-send double data to PC-6001
void P6_sndw_post(void)
{ 
	// PORTC = 0b11111111;	
	COMU_OUT_B(0b11111111);
	
	// DDRA  = 0b11111111;	// data out to P6
	COMU_DIR_A(0b11111111);
	
	// DDRC  = 0b00000000;	// data in from P6
	COMU_DIR_B(0b00000000);
}

// pre-receive double data to PC-6001
void P6_rcvw_pre(void)
{ 
	// PORTA = 0b11111111;
	COMU_OUT_A(0b11111111);
	
	// PORTC = 0b11111111;
	COMU_OUT_B(0b11111111);
	
	// DDRA  = 0b00000000;	// data in from P6
	COMU_DIR_A(0b00000000);
	
	// DDRC  = 0b00000000;	// data in from P6
	COMU_DIR_B(0b00000000);
	
	// PORTD |= 0b00100000;								// set RFD
	COMU_OUT_C(PIN1_bm);
	
	// while (bit_is_clear(PIND,0)) ;						// wait DAV set
	while (!(COMU_IN_C()&PIN4_bm)) ;

	// PORTD &= 0b11011111;								// reset RFD
	COMU_OUT_C(0);
}

// receive double data from PC-6001
void P6_rcvw(uint8_t *d1, uint8_t *d2, uint8_t *d3, uint8_t *d4)
{
	// while (bit_is_clear(PIND,0)) ;						// wait DAV set
	while (!(COMU_IN_C()&PIN4_bm)) ;
	
	//*d1 = PINC;
	*d1 = COMU_IN_B();
	//*d2 = PINA;
	*d2 = COMU_IN_A();
	// PORTD |= 0b01000000;								// set DAC
	COMU_OUT_C(PIN3_bm);
	
	// while (bit_is_set(PIND,0)) ;						// wait DAV reset
	while (COMU_IN_C()&PIN4_bm) ;
	
	//*d3 = PINC;
	*d3 = COMU_IN_B();	
	//*d4 = PINA;
	*d4 = COMU_IN_A();
	
	// PORTD &= 0b10111111;								// reset DAC
	COMU_OUT_C(PIN3_bm | 0);
}

// post-receive double data to PC-6001
void P6_rcvw_post(void)
{ 
	//PORTC = 0b11111111;
	COMU_OUT_C(0b11111111);
	
	//DDRA  = 0b11111111;	// data out to P6
	COMU_DIR_A(0b11111111);
	
	//DDRC  = 0b00000000;	// data in from P6
	COMU_DIR_B(0b00000000);
}

// initialize PC-6031 status
void pc6031_init(void)
{
	uint16_t i;

	pc6031.valid = 0;
	pc6031.error = 0;
	pc6031.mode = 0;
	for (i = 0;i != DRIVE_NUM;i ++) {
		pc6031.startSec[i] = 1;
		pc6031.endSec[i]   = 16;
		globalProtect[i] = 0;
	}
}

// process PC-6031 command
void pc6031_process(void)
{
	int16_t cnt = 0;

	sei();

	while (1) {
		uint8_t cmd;

		while (!(COMU_IN_C()&PIN7_bm)) {
			//SD_checkEject
			MOUNT_exchangeSD(0);

			int16_t new_cnt = TCD0_CNT;

			if (isFlashAir && ((new_cnt - cnt) < 0)) {
				LED2_ON;
				FLASHAIR_command();
				LED2_OFF;
			}
			cnt = new_cnt;
			
		};	// wait attention

		cli();
		P6_rcv(&cmd, 0);					// receive command
		sei();
		
		while (COMU_IN_C()&PIN7_bm) ;		// wait not attention

		cli();

		switch (cmd) {
		case 0:		// init
			pc6031.error = 0;
			pc6031.valid = 0;
			break;
		case 1:			// write
		case 17: 		// high speed write
		case 0x83:		// expand write
		case 0x84: 		// expand high speed write
		case 0x86: 		// double high speed write
		case 0x87: {	// expand double high speed write
			uint8_t lg, dr, trkupper,trklower, sec, i;
			uint16_t j;
			uint16_t trk;
			
			pc6031.error = 0;
			trkupper = 0;
			P6_rcv(&lg, 0); P6_rcv(&dr, 0);
			if (cmd==0x83 || cmd==0x84 || cmd==0x87) P6_rcv(&trkupper, 0);
			P6_rcv(&trklower, 0);
			trk  = (uint16_t)trklower;
			trk += (uint16_t)trkupper*0x100;
			P6_rcv(&sec, 0);

			if (!buffer2.sd6031.img[dr].startCluster) { pc6031.error = 1; break; }

			LED1_ON;			
				
			if (!(pow2(dr)&pc6031.mode)) trk<<=1;
			i = 0;
			do {
				if (buffer2.sd6031.img[dr].protect || globalProtect[dr]) {
					pc6031.error = 1;
				} else {	
					uint8_t err=0;
					uint32_t long_sector  = (uint32_t)(sec-pc6031.startSec[dr]);
					long_sector += (uint32_t)trk * (uint32_t)(pc6031.endSec[dr]-pc6031.startSec[dr]+1);

					FILE_writeBegin(&buffer2.sd6031.img[dr], long_sector, &err);
					if (err) {
						pc6031.error = 1;
					}
				}
				if (cmd==1 || cmd==0x83)
					for (j=0; j!=256; j++) {
						uint8_t err = 0;
						uint8_t d;
						
						P6_rcv(&d, 0);
						if (pc6031.error == 0) SPI_writeByte(d, &err);
					}
				else if (cmd==17 || cmd==0x84)
					for (j=0; j!=128; j++) {
						uint8_t err = 0;
						uint8_t d1, d2;
						
						P6_rcv(&d1, &d2);
						if (pc6031.error == 0) {
							SPI_writeByte(d1, &err);
							SPI_writeByte(d2, &err);
						}
					}			
				else {
					P6_rcvw_pre();
					for (j=0; j!=64; j++) {
						uint8_t d1, d2, d3, d4;
						
						P6_rcvw(&d1, &d2, &d3, &d4);
						if (pc6031.error == 0) {
							uint8_t err = 0;
							SPI_writeByte(d1, &err);
							SPI_writeByte(d2, &err);
							SPI_writeByte(d3, &err);
							SPI_writeByte(d4, &err);
						}
					}			
					P6_rcvw_post();
				}

				if (pc6031.error == 0) {
					uint8_t err = 0;
					for (j=0; j!=256; j++) SPI_writeByte(0, &err);
					FILE_writeEnd(&err);
				}
				if (sec == pc6031.endSec[dr]) {
					sec = pc6031.startSec[dr];
				} else {
					sec++;
				}
				i ++;
			} while (i != lg);

			LED1_OFF;
			break;
		}
		case 2:			// read
		case 0x85: {	// expand read
			uint8_t trkupper,trklower;
			trkupper = 0;
			P6_rcv(&pc6031.length, 0);
			P6_rcv(&pc6031.drive, 0);
			if (cmd==0x85) P6_rcv(&trkupper, 0);
			P6_rcv(&trklower, 0);
			pc6031.track  = (uint16_t)trklower;
			pc6031.track += (uint16_t)trkupper*0x100;
			P6_rcv(&pc6031.sector, 0);

			if (!buffer2.sd6031.img[pc6031.drive].valid) { pc6031.error=1; break; }

			LED2_ON;
			if (!(pow2(pc6031.drive)&pc6031.mode)) pc6031.track<<=1;
			pc6031.valid = 1;

			pc6031.error = 0;
			//LED = 0;
			//PORTB = ((PORTB&0b11111100) | LED);
			LED2_OFF;
			break;
		}
		case 3:			// transfer read data
		case 18: 		// high speed transfer read data
		case 0x88: {	// double high speed transfer read data
			uint8_t i;
			uint16_t j;

			pc6031.error = 0;
			if ((!pc6031.valid) || (!buffer2.sd6031.img[pc6031.drive].startCluster)) { pc6031.error=1; break; }

			LED2_ON;
			i = 0;
			do {
				uint8_t err = 0;
				uint32_t long_sector  = (uint32_t)(pc6031.sector-pc6031.startSec[pc6031.drive]);
				long_sector += (uint32_t)pc6031.track * (uint32_t)(pc6031.endSec[pc6031.drive]-pc6031.startSec[pc6031.drive]+1);

				FILE_readBegin(&buffer2.sd6031.img[pc6031.drive], long_sector, &err);

				if (err) {
					pc6031.error = 1;
				}
				
				if (cmd==3) for (j=0; j!=256; j++) {
					uint8_t d;
					if (pc6031.error == 0) {
						d = SPI_readByte(&err);
					} else {
						d = 0xff;
					}
					
					P6_snd(d, 0);
				} else if (cmd==18) for (j=0; j!=128; j++) {
					uint8_t d1;
					uint8_t d2;
					if (pc6031.error == 0) {
						d1 = SPI_readByte(&err);
						d2 = SPI_readByte(&err);
					} else {
						d1 = 0xff;
						d2 = 0xff;
					}
					
					P6_snd(d1, d2);
				} else {
					P6_sndw_pre();
					for (j=0; j!=64; j++) {
						uint8_t d1;
						uint8_t d2;
						uint8_t d3;
						uint8_t d4;
						if (pc6031.error == 0) {
							d1 = SPI_readByte(&err);
							d2 = SPI_readByte(&err);
							d3 = SPI_readByte(&err);
							d4 = SPI_readByte(&err);
						} else {
							d1 = 0xff;
							d2 = 0xff;
							d3 = 0xff;
							d4 = 0xff;
						}
						P6_sndw(d1, d2, d3, d4);
					}
					P6_sndw_post();
				}
				if (pc6031.error == 0) {
					for (j=0; j!=256; j++) SPI_readByte(&err);
					FILE_readEnd(&err);
				}
				if (pc6031.sector == pc6031.endSec[pc6031.drive]) {
					pc6031.sector = pc6031.startSec[pc6031.drive];
				} else {
					pc6031.sector++;
				}
				i ++;
			} while (i != pc6031.length);
			pc6031.valid = 0;

			LED2_OFF;
			break;
		}
		case 5:	{	// format
			uint8_t trk, sec;
			uint8_t drv, is1d;
			uint16_t j;
				
			pc6031.error = 0;
			P6_rcv(&drv, 0);	
			if (buffer2.sd6031.img[drv].protect || globalProtect[drv] || !buffer2.sd6031.img[drv].startCluster) { pc6031.error = 1; break; }
			is1d = (!(pow2(drv)&pc6031.mode));

			LED1_ON;
			// initialize the area with 0xff
			for (trk=0; trk!=80; trk++) {
				if (is1d && (trk&1) ) continue;
				for (sec=1; sec!=17; sec++) {
					uint8_t err=0;
					uint32_t long_sector  = (uint32_t)(sec-pc6031.startSec[drv]);
					long_sector += (uint32_t)trk * (uint32_t)(pc6031.endSec[drv]-pc6031.startSec[drv]+1);
					FILE_writeBegin(&buffer2.sd6031.img[drv], long_sector, &err);
					
					if (err) {
						pc6031.error = 1;
						break;
					}

					// format with 0xff
					for (j=0; j!=256; j++) SPI_writeByte(0xff, &err);
					// dummy 256 bytes
					for (j=0; j!=256; j++) SPI_writeByte(0, &err);
					FILE_writeEnd(&err);
					if (err) {
						pc6031.error = 1;
						break;
					}
				}
				if (pc6031.error == 1) {
					break;
				}
			}

			LED1_OFF;
			break;
		}
		case 6: {		// command status
			uint8_t st = ((0b10000000 | (pc6031.valid<<6) | pc6031.error));

			P6_snd(st, 0);
			pc6031.error = 0;
			break;
		}
		case 7: {	// drive status
			const uint8_t status =
				(DRIVE_NUM==1)?0b00011111:((DRIVE_NUM==2)?0b00111111:
				((DRIVE_NUM==3)?0b01111111:0b11111111));

			P6_snd(status, 0);
			pc6031.error = 0;
			break;
		}
		case 11: {	// transfer FDD's memory data to PC-6001
					// only for address 7EFh (return EFh) & 75Fh (return 77h)
			uint16_t i;
			uint8_t AH, AL, DH, DL;

			P6_rcv(&AH, 0); P6_rcv(&AL, 0); P6_rcv(&DH, 0); P6_rcv(&DL, 0);
			P6_snd((AH==7)?((AL==0x5f)?0x77:((AL==0xef)?0xef:0xff)):0xff, 0);
			for (i = 1; i < ((((short)DH)<<8)|DL); i++) P6_snd(0xff, 0);
			pc6031.error = 0;
			break;
		}
		case 20: {
			uint8_t drv;
			
			P6_rcv(&drv, 0);
			P6_snd(buffer2.sd6031.img[drv].protect?0b01000000:0,0);
			pc6031.error = 0;
			break;
		}
		case 23: {	// set mode (1D or 2D)
			uint8_t d;
			P6_rcv(&d, 0);
			pc6031.mode =(d&0b00001111);
			pc6031.error = 0;
			break;
		}
		case 24: {	// ask mode
			P6_snd(pc6031.mode, 0);
			pc6031.error = 0;
			break;
		}
		case 0x80: {	// set SECTOR data
			uint8_t drv;
			P6_rcv(&drv, 0);
			P6_rcv(&(pc6031.startSec[drv]), 0);
			P6_rcv(&(pc6031.endSec[drv])  , 0);
			pc6031.error = 0;
			break;
		}
		case 0x81: {	// ask SECTOR data
			uint8_t drv;
			P6_rcv(&drv, 0);
			P6_snd(pc6031.startSec[drv], 0);
			P6_snd(pc6031.endSec[drv]  , 0);
			pc6031.error = 0;
			break;
		}
		case 0x82: {	// report DISK IMAGE length
			uint8_t drv;
			uint8_t tmp;
			uint32_t tmplen;

			P6_rcv(&drv, 0);
			tmplen = buffer2.sd6031.img[drv].length;
			tmp = tmplen / 0x1000000;
			P6_snd(tmp,0);
			tmplen = tmplen - tmp * 0x1000000;
			tmp = tmplen / 0x10000;
			P6_snd(tmp,0);
			tmplen = tmplen - tmp * 0x10000;
			tmp = tmplen / 0x100;
			P6_snd(tmp,0);
			tmplen = tmplen - tmp * 0x100;
			tmp = tmplen;
			P6_snd(tmp,0);

			pc6031.error = 0;
			break;
		}
		case 0xFD: { // send P31 image names to P6
			uint8_t drv;

			P6_rcv(&drv, 0);

			LED2_ON;
			sendFileNames(drv);
			pc6031.error = 0;
			LED2_OFF;
			break;
		}
		case 0xFE: { // change image
			uint8_t drv, i;
			char name[23]="D                  P31";

			pc6031.error = 1;
			P6_rcv(&drv, 0);
			name[1] = ('0'+drv);
			LED2_ON;
			for (i=0; i!=8; i++)
				P6_rcv((uint8_t *)&(name[i+11]), 0);
			uint8_t err = 0;
			FILE_openAbs(&buffer2.sd6031.img[drv], name, 1, &err);
			if (!err) {
				pc6031.error = 0;
				
				struct FILE tempImg;
				memcpy(&name[11],"2D         ",11);
				err = 0;
				FILE_openAbs(&tempImg, name, 1, &err);
				if (!err) pc6031.mode |= pow2(drv);
			}
			LED2_OFF;
			break;
		}
		}
		sei();
	}
}

void SD6031_init(void)
{
	COMU_init();
	COMU_OUT_A(0b00000000);
	COMU_OUT_C(0b00000000);
	COMU_DIR_A(0b11111111);	// data out to P6
	COMU_DIR_B(0b00000000);	// data in from P6
	COMU_DIR_C(0b00001111);	// ctrl in from P6 (ATN,DAC,RFD,DAV)
	// ctrl out to P6 (ATN,DAC,RFD,DAV),
	
	PORTE.INT0MASK = PIN0_bm;
	PMIC.CTRL = PMIC_LOLVLEN_bm;
	PORTE.INTCTRL = PORT_INT0LVL_LO_gc;
}

void SD6031_run(void)
{
	// TCD0 about 1 sec used for FlashAir
	TCD0.PER = 31250;
	TCD0.CTRLA = TC0_CLKSEL2_bm|TC0_CLKSEL1_bm|TC0_CLKSEL0_bm;			// prescaler : clk/1024
	TCD0.INTCTRLA = 0;

	pc6031_init();
		
	sei();

	pc6031_process();
}

// reset!
ISR(PORTE_INT0_vect)
{
#ifdef AIRDISP	
	LCD_locate(0,0);
	LCD_print("reset",5);
#endif
	_delay_ms(100);

	CPU_CCP=CCP_IOREG_gc;
	RST.CTRL=RST_SWRST_bm;
	while (1);
}